@@ -132,7 +132,13 @@ class GWCSAPIMixin(BaseHighLevelWCS, BaseLowLevelWCS):
         be returned in the ``(x, y)`` order, where for an image, ``x`` is the
         horizontal coordinate and ``y`` is the vertical coordinate.
         """
-        world_arrays = self._add_units_input(world_arrays, self.backward_transform, self.output_frame)
+        try:
+            backward_transform = self.backward_transform
+            world_arrays = self._add_units_input(world_arrays,
+                                                 backward_transform,
+                                                 self.output_frame)
+        except NotImplementedError:
+            pass
 
         result = self.invert(*world_arrays, with_units=False)
 
@@ -147,10 +153,14 @@ class GWCSAPIMixin(BaseHighLevelWCS, BaseLowLevelWCS):
         `~BaseLowLevelWCS.pixel_to_world_values`). The indices should be
         returned as rounded integers.
         """
-        result = self.world_to_pixel_values(*world_arrays)
-        if self.pixel_n_dim != 1:
-            result = result[::-1]
-        return result
+        results = self.world_to_pixel_values(*world_arrays)
+        if self.pixel_n_dim == 1:
+            results = (results,)
+        else:
+            results = results[::-1]
+
+        results = tuple(utils._toindex(result) for result in results)
+        return results[0] if self.pixel_n_dim == 1 else results
 
     @property
     def array_shape(self):
@@ -313,7 +323,6 @@ class GWCSAPIMixin(BaseHighLevelWCS, BaseLowLevelWCS):
         Convert world coordinates to pixel values.
         """
         result = self.invert(*world_objects, with_units=True)
-
         if self.input_frame.naxes > 1:
             first_res = result[0]
             if not utils.isnumerical(first_res):
